"""Data structures required for representing the contents of a McCode instr file"""
from __future__ import annotations

from io import StringIO
from dataclasses import dataclass, field
from typing import Union
from ..common import InstrumentParameter, MetaData, parameter_name_present, RawC, blocks_to_raw_c, Expr, Value
from ..reader import Registry
from .instance import Instance
from .group import Group
from loguru import logger


@dataclass
class Instr:
    """Intermediate representation of a McCode instrument

    Read from a .instr file -- possibly including more .comp and .instr file sources
    For output to a runtime source file
    """
    name: str = None  # Instrument name, e.g. {name}.instr (typically)
    source: str = None  # Instrument *file* name
    parameters: tuple[InstrumentParameter, ...] = field(default_factory=tuple)  # runtime-set instrument parameters
    metadata: tuple[MetaData, ...] = field(default_factory=tuple)  # metadata for use by simulation consumers
    components: tuple[Instance, ...] = field(default_factory=tuple)  #
    included: tuple[str, ...] = field(default_factory=tuple)  # names of included instr definition(s)
    user: tuple[RawC, ...] = field(default_factory=tuple)  # struct members for _particle
    declare: tuple[RawC, ...] = field(default_factory=tuple)  # global parameters used in component trace
    initialize: tuple[RawC, ...] = field(default_factory=tuple)  # initialization of global declare parameters
    save: tuple[RawC, ...] = field(default_factory=tuple)  # statements executed after TRACE to save results
    final: tuple[RawC, ...] = field(default_factory=tuple)  # clean-up memory for global declare parameters
    groups: dict[str, Group] = field(default_factory=dict)
    flags: tuple[str, ...] = field(default_factory=tuple)  # (C) flags needed for compilation of the (translated) instrument
    registries: tuple[Registry, ...] = field(default_factory=tuple)  # the registries used by the reader to populate this

    def to_file(self, output=None, wrapper=None):
        if output is None:
            output = StringIO()
        if wrapper is None:
            from mccode_antlr.common import TextWrapper
            wrapper = TextWrapper(width=120)
        print(wrapper.start_block_comment(f'Instrument {self.name}'), file=output)
        print(wrapper.line('Instrument:', [self.name or 'None']), file=output)
        print(wrapper.line('Source:', [self.source or 'None']), file=output)
        print(wrapper.line('Contains:', [f'"%include {include}"' for include in self.included]), file=output)
        print(wrapper.line('Registries:', [registry.name for registry in self.registries]), file=output)
        for registry in self.registries:
            registry.to_file(output=output, wrapper=wrapper)
        print(wrapper.end_block_comment(), file=output)

        instr_parameters = wrapper.hide(', '.join(p.to_string(wrapper=wrapper) for p in self.parameters))
        first_line = wrapper.line('DEFINE INSTRUMENT', [f'{self.name}({instr_parameters})'])
        print(first_line, file=output)

        for metadata in self.metadata:
            metadata.to_file(output=output, wrapper=wrapper)
        if self.flags:
            print(wrapper.quoted_line('DEPENDENCY ', list(self.flags)), file=output)

        if self.declare:
            print(wrapper.block('DECLARE', _join_rawc_tuple(self.declare)), file=output)
        if self.user:
            print(wrapper.block('USERVARS', _join_rawc_tuple(self.user)), file=output)
        if self.initialize:
            print(wrapper.block('INITIALIZE', _join_rawc_tuple(self.initialize)), file=output)

        print(wrapper.start_list('TRACE'), file=output)
        for instance in self.components:
            print(wrapper.start_list_item(), file=output)
            instance.to_file(output, wrapper)
            print(wrapper.end_list_item(), file=output)
        if self.save:
            print(wrapper.block('SAVE', _join_rawc_tuple(self.save)), file=output)
        if self.final:
            print(wrapper.block('FINALLY', _join_rawc_tuple(self.final)), file=output)
        print(wrapper.end_list('END'), file=output)

    def to_string(self, wrapper):
        from io import StringIO
        output = StringIO()
        self.to_file(output, wrapper)
        return output.getvalue()

    def __str__(self):
        from mccode_antlr.common import TextWrapper
        return self.to_string(TextWrapper())

    def _my_repr_html_(self):
        from mccode_antlr.common import HTMLWrapper
        wrapper = HTMLWrapper(hider='hider', hidden='hidden')
        output = StringIO()
        self.to_file(output=output, wrapper=wrapper)
        body = output.getvalue()
        style = """
        <style> 
        .hider {cursor: pointer; user-select: none;}
        .hider::before {content: "...";}
        .hidden-before::before {display: none;}
        .hidden {display: none;}
        .active {display: block;}
        </style>
        """
        script = """
        <script>
        var toggler = document.getElementsByClassName("hider");
        var i;
        for (i = 0; i < toggler.length; i++) {
            toggler[i].addEventListener("click", function() {
                var id = this.getAttribute('data-id');
                this.parentElement.querySelector(`.hidden[id=${id}]`).classList.toggle("active");
                this.classList.toggle("hidden-before");
            });
        }
        </script>
        """
        html = f'<html><head><title>{self.name}</title>{style}</head><body>{body}{script}</body></html>'
        return html

    def add_component(self, a: Instance):
        if any(x.name == a.name for x in self.components):
            raise RuntimeError(f"A component instance named {a.name} is already present in the instrument")
        self.components += (a,)

    def add_parameter(self, a: InstrumentParameter, ignore_repeated=False):
        if not parameter_name_present(self.parameters, a.name):
            self.parameters += (a,)
        elif not ignore_repeated:
            raise RuntimeError(f"An instrument parameter named {a.name} is already present in the instrument")

    def get_parameter(self, name, default=None):
        if parameter_name_present(self.parameters, name):
            for parameter in self.parameters:
                if name == parameter.name:
                    return parameter
        return default

    def has_parameter(self, name):
        return parameter_name_present(self.parameters, name)

    def last_component(self, count: int = 1, removable_ok: bool = True):
        if len(self.components) < count:
            raise RuntimeError(f"Only {len(self.components)} components defined -- can not go back {count}.")
        if removable_ok:
            return self.components[-count]
        fixed = [comp for comp in self.components if not comp.removable]
        if len(fixed) < count:
            for comp in self.components:
                logger.info(f'{comp.name}')
            raise RuntimeError(f"Only {len(fixed)} fixed components defined -- can not go back {count}.")
        return fixed[-count]

    def get_component(self, name: str):
        if name == 'PREVIOUS':
            return self.components[-1]
        for comp in self.components:
            if comp.name == name:
                return comp
        raise RuntimeError(f"No component instance named {name} defined.")

    def has_component_named(self, name: str):
        return any(comp.name == name for comp in self.components)

    def get_component_names_by_category(self, category: str):
        """Find all component instance names for a given category.

        Note:
            The category of an instance is determined by its type, and is either
            - set inside the .comp file by a 'CATEGORY <category>' directive, or
            - *guessed* by the reader based on the path to the .comp file.
            The second behaviour is to match McStasScript/McCode-3, which does not work for user-defined components.
        """
        return [inst.name for inst in self.components if category in inst.type.category]

    def add_included(self, name: str):
        self.included += (name,)

    def DEPENDENCY(self, *strings):
        self.flags += strings

    def USERVARS(self, *blocks):
        self.user += blocks_to_raw_c(*blocks)

    def DECLARE(self, *blocks):
        self.declare += blocks_to_raw_c(*blocks)

    def INITIALIZE(self, *blocks):
        self.initialize += blocks_to_raw_c(*blocks)

    def SAVE(self, *blocks):
        self.save += blocks_to_raw_c(*blocks)

    def FINALLY(self, *blocks):
        self.final += blocks_to_raw_c(*blocks)

    def add_metadata(self, m: MetaData):
        if any([x.name == m.name for x in self.metadata]):
            self.metadata = tuple([x for x in self.metadata if x.name != m.name])
        self.metadata += (m,)

    def determine_groups(self):
        for id, inst in enumerate(self.components):
            if inst.group:
                if inst.group not in self.groups:
                    self.groups[inst.group] = Group(inst.group, len(self.groups))
                self.groups[inst.group].add(id, inst)

    def component_types(self):
        # # If component order is unimportant, we can use a set:
        # return set(inst.type for inst in self.components)
        # For comparison with the C code generator, we must keep the order of component definitions
        return list(dict.fromkeys([inst.type for inst in self.components]))

    def collect_metadata(self):
        """Component definitions and instances can define metadata too, collect it all together here"""
        # Metadata defined in an instance overrides that defined in a component.
        # Metadata defined for an instrument is added to the collected list
        return tuple(m for inst in self.components for m in inst.collect_metadata()) + self.metadata

    def _getpath(self, filename: str):
        from pathlib import Path
        for registry in self.registries:
            if registry.known(filename):
                return registry.path(filename).absolute().resolve()
        return Path()

    def _replace_env_getpath_cmd(self, flags: str):
        """Replace CMD, ENV, and GETPATH directives from a flag string"""

        # Mimics McCode-3/tools/Python/mccodelib/cflags.py:evaluate_dependency_str
        #
        def getpath(chars):
            return self._getpath(chars).as_posix()

        def eval_cmd(chars):
            from subprocess import run, CalledProcessError
            try:
                proc = run(chars, check=True, shell=True, capture_output=True, text=True)
                output = proc.stdout
            except CalledProcessError as error:
                raise RuntimeError(f"Calling {chars} resulted in error {error}")
            output = [line.strip() for line in output.splitlines() if line.strip()]
            if len(output) > 1:
                raise RuntimeError(f"Calling {chars} produced more than one line of output")
            return output[0] if output else ''

        def eval_env(chars):
            from os import environ
            return environ.get(chars, '')

        def replace(chars, start, replacer):
            if start not in chars:
                return chars
            before, after = chars.split(start, 1)
            if '(' != after[0]:
                raise ValueError(f'Missing opening parenthesis in dependency string after {start}')
            if ')' not in after:
                raise ValueError(f'Missing closing parenthesis in dependency string after {start}')
            dep, after = after[1:].split(')', 1)
            if start in dep:
                raise ValueError(f'Nested {start} in dependency string')
            return before + replacer(dep) + replace(after, start, replacer)

        for key, worker in zip(['ENV', 'GETPATH', 'CMD'], [eval_env, getpath, eval_cmd]):
            flags = replace(flags, key, worker)

        return flags

    def _replace_keywords(self, flag):
        from mccode_antlr.config import config
        from re import sub, findall
        if '@NEXUSFLAGS@' in flag:
            flag = sub(r'@NEXUSFLAGS@', config['flags']['nexus'].as_str_expanded(), flag)
        if '@MCCODE_LIB@' in flag:
            print(f'The instrument {self.name} uses @MCCODE_LIB@ dependencies which no longer work.')
            print('Expect problems at compilation.')
            flag = sub('@MCCODE_LIB@', '.', flag)
        general_re = r'@(\w+)@'
        for replace in findall(general_re, flag):
            # Is this replacement something like XXXFLAGS?
            if replace.lower().endswith('flags') and replace.lower()[:-5] in config['flags']:
                flag = sub(f'@{replace}@', config['flags'][replace.lower()[:-5]].as_str_expanded(), flag)
            elif replace.lower().endswith('flags'):
                # Punt, and just replace with the lowercase version of the keyword -- hopefully this was intended
                flag = sub(f'@{replace}@', f'-l{replace.lower()[:-5]}', flag)
            else:
                logger.warning(f'Unknown keyword @{replace}@ in dependency string')
        return flag

    def decoded_flags(self) -> list[str]:
        # Each 'flag' in self.flags is from a single instrument component DEPENDENCY, and might contain duplicates:
        # If we accept that white space differences matter, we can deduplicate the strings 'easily'
        unique_flags = set(self.flags)
        # The dependency strings are allowed to contain any of
        #       '@NEXUSFLAGS@', @MCCODE_LIB@, CMD(...), ENV(...), GETPATH(...)
        # each of which should be replaced by ... something. Start by replacing the 'static' (old-style) keywords
        replaced_flags = [self._replace_keywords(flag) for flag in unique_flags]
        # Then use the above decoder method to replace any instances of CMD, ENV, or GETPATH
        return [self._replace_env_getpath_cmd(flag) for flag in replaced_flags]

    def copy(self, first=0, last=-1):
        """Return a copy of this instrument, optionally with only a subset of components"""
        from copy import deepcopy
        copy = Instr(self.name, self.source)
        copy.parameters = tuple(x for x in self.parameters)
        copy.metadata = tuple(x.copy() for x in self.metadata)
        if last < 0:
            last += 1 + len(self.components)
        copy.components = tuple(x.copy() for x in self.components[first:last])
        copy.included = tuple(x for x in self.included)
        copy.user = tuple(x.copy() for x in self.user)
        copy.declare = tuple(x.copy() for x in self.declare)
        copy.initialize = tuple(x.copy() for x in self.initialize)
        copy.save = tuple(x.copy() for x in self.save)
        copy.final = tuple(x.copy() for x in self.final)
        copy.groups = {k: v.copy() for k, v in self.groups.items()}
        copy.flags = tuple(x for x in self.flags)
        copy.registries = tuple(x for x in self.registries)
        return copy

    def split(self, at, remove_unused_parameters=False):
        """Produces two instruments, both containing the indicated component

        Parameters:
        -----------
        after: Union[Instance, str]
            A component instance or the _name_ of a component instance at which to split the instrument.
            The instance or one with a matching name _must_ be in the instrument, and should probably be an Arm.
        remove_unused_parameters: bool
            If True, any Instrument parameters which do not appear in instance definitions or code blocks is not
            included in the output instruments

        Return:
        -------
        tuple[Instr, Instr]
            The first Instr has components up to and including the split-point.
            The second Instr has components starting from the split-point.
        """
        if isinstance(at, Instance):
            index = self.components.index(at)
        elif isinstance(at, str):
            index = [i for i, x in enumerate(self.components) if x.name == at]
            if len(index) != 1:
                raise RuntimeError(f'Can only split an instrument after a single component, "{at}" matches {index}')
            index = index[0]
        else:
            raise RuntimeError('Can only split an instrument after a component or component name')
        first = self.copy(last=index + 1)
        first.name = self.name + '_first'
        if first.check_instrument_parameters(remove=remove_unused_parameters) and not remove_unused_parameters:
            logger.warning(f'Instrument {first.name} has unused instrument parameters')

        second = self.copy(first=index)
        second.name = self.name + '_second'
        # remove any dangling component references and re-reference into the new instrument's components:
        for instance in second.components:
            at_rel = instance.at_relative[1]
            rot_rel = instance.rotate_relative[1]
            if isinstance(at_rel, Instance):
                if second.has_component_named(at_rel.name):
                    instance.at_relative = instance.at_relative[0], second.get_component(at_rel.name)
                else:
                    instance.at_relative = instance.orientation.position(), None
            if isinstance(rot_rel, Instance):
                if second.has_component_named(rot_rel.name):
                    instance.rotate_relative = instance.rotate_relative[0], second.get_component(rot_rel.name)
                else:
                    instance.rotate_relative = instance.orientation.angles(), None
        if second.check_instrument_parameters(remove=remove_unused_parameters) and not remove_unused_parameters:
            logger.info(f'Instrument {second.name} has unused instrument parameters')

        return first, second

    def make_instance(self, name, component, at_relative=None, rotate_relative=None, orientation=None,
                      parameters=None, group=None, removable=False):
        if parameters is None:
            parameters = tuple()
        if any(x.name == name for x in self.components):
            raise RuntimeError(f"An instance named {name} is already present in the instrument")
        if isinstance(component, str):
            from ..reader import Reader
            reader = Reader(registries=list(self.registries))
            component = reader.get_component(component)
            self.flags += tuple(reader.c_flags)
        self.components += (Instance(name, component, at_relative, rotate_relative, orientation,
                                     parameters, group, removable),)

    def mcpl_split(self,
                   after,
                   filename=None,
                   output_component=None,
                   output_parameters=None,
                   input_component=None,
                   input_parameters=None,
                   remove_unused_parameters=False
                   ):
        from ..common import ComponentParameter
        from ..common import Expr, Value, ObjectType
        from .orientation import Vector, Angles
        if filename is None:
            filename = self.name + '.mcpl'
        if filename[0] != '"' or filename[-1] != '"':
            filename = '"' + filename + '"'

        filename_parameter = ComponentParameter('filename', Expr(Value('mcpl_filename', object_type=ObjectType.parameter)))
        first, second = self.split(after, remove_unused_parameters=remove_unused_parameters)
        mcpl_filename = InstrumentParameter.parse(f'string mcpl_filename = {filename}')
        first.add_parameter(mcpl_filename)
        second.add_parameter(mcpl_filename)

        fc = first.components[-1]
        if fc.type.name != 'Arm':
            logger.info(f'Component {after} is a {fc.type.name} instead of an Arm -- using MCPL file may cause problems')

        if output_component is None:
            output_component = 'MCPL_output'
        if output_parameters is None:
            output_parameters = (filename_parameter,)
        elif not any(p.name == 'filename' for p in output_parameters):
            output_parameters = (filename_parameter,) + output_parameters
        # remove the last component, since we're going to re-use its name:
        first.components = first.components[:-1]
        # automatically adds the component at the end of the list:
        first.make_instance(fc.name, output_component, fc.at_relative, fc.rotate_relative, fc.orientation,
                            output_parameters)

        if input_component is None:
            input_component = 'MCPL_input'
        if input_parameters is None:
            input_parameters = (filename_parameter,)
        elif not any(p.name == 'filename' for p in input_parameters):
            input_parameters = (filename_parameter,) + input_parameters
        if not any(p.name == 'verbose' for p in input_parameters):
            input_parameters = (ComponentParameter('verbose', Expr.float(0)),) + input_parameters
        # # the MCPL input component _is_ the origin of its simulation, but must be placed relative to other components.
        # # so we need the *absolute* position and orientation of the removed component:
        # abs_at_rel = fc.orientation.position(), None
        # abs_rot_rel = fc.orientation.angles(), None

        # the split at component in the second instrument should have already been converted to absolute-positioning:
        sc = second.components[0]
        if sc.at_relative[1] is not None or sc.rotate_relative[1] is not None:
            logger.error("The split-at point should be positioned absolutely in the second instrument")
        # remove the first component before adding an-equal named one:
        second.components = second.components[1:]
        second.make_instance(sc.name, input_component, sc.at_relative, sc.rotate_relative, parameters=input_parameters)
        # move the newly added component to the front of the list:
        second.components = (second.components[-1],) + second.components[:-1]

        return first, second

    def parameter_used(self, name: str):
        """Check that an instrument parameter is used in the instrument"""
        for instance in self.components:
            if instance.parameter_used(name):
                return True
        for section in (self.declare, self.initialize, self.save, self.final):
            for block in section:
                # A more complex check would see if the use itself leads to a parameter being used, but
                # that would be language dependent and probably not worth the effort.
                if name in block:
                    return True
        return False

    def check_instrument_parameters(self, remove=False):
        """Check that all instrument parameters are used in the instrument, and optionally remove any that are not

        Returns
        -------
        int
            The number of unused instrument parameters
        """
        names = [p.name for p in self.parameters]
        used = [self.parameter_used(p.name) for p in self.parameters]
        if not all(used):
            logger.info(f'The following instrument parameters are not used in the instrument: '
                     f'{", ".join([n for n, u in zip(names, used) if not u])}')
            if remove:
                self.parameters = tuple(p for i, p in enumerate(self.parameters) if used[i])
                logger.info(f'Removed unused instrument parameters; {len(self.parameters)} remain')
        return len(used) - sum(used)

    def verify_instance_parameters(self):
        """Check that all instance parameters are of the expected type, and that identifiers which match
        instrument parameter names are flagged as such"""
        for instance in self.components:
            instance.verify_parameters(self.parameters)

    def check_expr(self, expr: int | float | str | Expr | Value):
        if not isinstance(expr, Expr):
            expr = Expr.best(expr)
        # check whether the expression contains any identifiers which are actually InstrumentParameters
        expr.verify_parameters([x.name for x in self.parameters])
        # We then verify that no as-of-yet undefined identifiers exist, but can't in case they're defined in
        # an initalize or share block
        return expr


def _join_rawc_tuple(rawc_tuple: tuple[RawC]):
    return '\n'.join([str(rc) for rc in rawc_tuple])



from __future__ import annotations

from antlr4 import InputStream
from antlr4.tree.Tree import ParseTree
from antlr4.error.ErrorListener import ErrorListener


def parse(
        stream: InputStream,
        entry_rule_name: str,
        error_listener: ErrorListener | None = None
) -> ParseTree:
    from .McCompParser import McCompParser
    from .McCompLexer import McCompLexer
    from .mccode_parse import parse as p
    return p(McCompLexer, McCompParser, stream, entry_rule_name, error_listener)



from __future__ import annotations

from importlib.metadata import PackageNotFoundError

from loguru import logger
from enum import Enum
from functools import cache

@cache
def antlr4_version(version: str | None = None):
    """Query maven for the most recent antlr4 version.

    Note
    ----
    This method copies part of the antlr4-tools argument parser to allow
    overriding any discovery by specifying an environment variable
    `$ANTLR4_TOOLS_ANTLR_VERSION`. If that environment variable is unset,
    maven is queried for the most up-to-date antlr4 version. If the query
    times out, the local maven folders are checked to find a suitable version
    (presumably the most recent one there).
    Since the maven query takes time and could time out, and we only want
    one version for all generated files, this method is cached to always
    give the same output during a single runtime.
    """
    from os import environ
    from antlr4_tool_runner import latest_version
    return version or environ.get("ANTLR4_TOOLS_ANTLR_VERSION") or latest_version()


def antlr4_runtime_version():
    """Retrieve the ANTLR4 version used by the available antlr4-python3-runtime"""
    from importlib import metadata
    try:
        return metadata.metadata('antlr4-python3-runtime').get('version')
    except PackageNotFoundError:
        logger.warning(f'The ANTLR4 Python runtime must match the ANTLR4 version, but is not installed')
        logger.info(f'Install antlr4-python3-runtime=={antlr4_version()} to use the generated files')
    return None


class Target(Enum):
    python = 0
    cpp = 1
    def __str__(self):
        if self == Target.python:
            return 'Python3'
        elif self == Target.cpp:
            return 'Cpp'
        raise ValueError(f'Unknown target {self}')


class Feature(Enum):
    listener = 0
    visitor = 1


def rebuild_language(grammar_file,
                     target: Target,
                     features: list[Feature],
                     verbose=False,
                     output=None,
                     dryrun=False,
                     version: str | None = None,
                     ):
    from pathlib import Path
    from subprocess import Popen, PIPE
    from antlr4_tool_runner import initialize_paths, install_jre_and_antlr

    if not isinstance(grammar_file, Path):
        grammar_file = Path(grammar_file)

    args =[
        f'-Dlanguage={target}',
        '-visitor' if Feature.visitor in features else '-no-visitor',
        '-listener' if Feature.listener in features else '-no-listener',
        '-o', output or str(grammar_file.parent / str(target)),
        str(grammar_file)
    ]

    if verbose:
        logger.info(f'Building language files for {grammar_file}')

    # The following copies the implementation of antlr4_tool_runner.tool, which pulls `args` from the system argv list
    # Setup:
    initialize_paths()

    if dryrun:
        return

    jar, java = install_jre_and_antlr(antlr4_version(version))
    # Call antlr4
    p = Popen([java, '-cp', jar, 'org.antlr.v4.Tool'] + args, stdout=PIPE, stderr=PIPE)
    out, err = [x.decode('UTF-8') for x in p.communicate()]
    if err:
        print(err, end='')
    if out:
        print(out, end='')


def language_missing_or_outdated(grammar_file, newest, features, path,
                                 verbose=False,
                                 dryrun=False,
                                 version: str | None = None,
                                 ):
    from pathlib import Path
    import re
    if not isinstance(grammar_file, Path):
        grammar_file = Path(grammar_file)

    if not grammar_file.exists():
        raise RuntimeError(f"The grammar file {grammar_file} does not exist.")

    lexer_file = path.joinpath(f'{grammar_file.stem}Lexer.py')
    parser_file = path.joinpath(f'{grammar_file.stem}Parser.py')
    listener_file = path.joinpath(f'{grammar_file.stem}Listener.py')
    visitor_file = path.joinpath(f'{grammar_file.stem}Visitor.py')

    generated_files = (lexer_file, parser_file)
    if Feature.visitor in features:
        generated_files += (visitor_file, )
    if Feature.listener in features:
        generated_files += (listener_file, )

    if not all(x.exists() for x in generated_files):
        if verbose:
            logger.info(f'Not all language files exist for {grammar_file}')
        return True

    if any(x.stat().st_mtime < newest for x in generated_files):
        if verbose:
            logger.info(f'Not all language files up-to-date for {grammar_file}')
        return True

    # Finally check if the antlr4 version of the generated file is as requested
    version = antlr4_version(version)
    # some antlr components use `self.checkVersion("{hard coded version string}")
    # to verify that they correspond to the same version as the runtime.
    r_checkversion = r'checkVersion\(\"(?P<version>[0-9]+\.[0-9]+\.[0-9]+)\"\)'
    # others (all?) have a comment string on their first line with ANTLR {version}
    r_antlr_version = r'ANTLR (?P<version>[0-9]+\.[0-9]+\.[0-9]+)'
    for file in generated_files:
        with file.open('r') as f:
            contents = f.read()
            checkversion_matches = re.findall(r_checkversion, contents, re.MULTILINE)
            antlr_version_matches = re.findall(r_antlr_version, contents, re.MULTILINE)
        if any(v != version for v in checkversion_matches + antlr_version_matches):
            if verbose:
                logger.info(f'Not all language files match requested ANTLR version {version}')
            return True

    if verbose:
        logger.info(f'Language files for {grammar_file} are up-to-date')
    return False


# def rebuild_speedy_language(grammar_file, features: list[Feature], output: Path, verbose=False):
#     from speedy_antlr_tool import generate
#
#     cpp_output_dir = grammar_file.parent / str(Target.cpp)
#
#     rebuild_language(grammar_file, Target.cpp, features, verbose=verbose, output=cpp_output_dir)
#     rebuild_language(grammar_file, Target.python, features, verbose=verbose, output=output)
#
#     py_parser_path = output / f'{grammar_file.stem}Parser.py'
#     generate(str(py_parser_path), cpp_output_dir)


def ensure_language_up_to_date(
        grammar: str,
        *,
        target: Target,
        features: list[Feature],
        deps=None,
        **kwargs
):
    """Ensure the ANTLR parsed language files are up-to-date."""
    from pathlib import Path
    # This file is at src/grammar/builder.py, it's parent is src/grammar
    # which the grammar defining files are under
    grammar_file = Path(__file__).parent / f'{grammar}.g4'
    # and we want to put Python files under src/mccode_antlr/grammar
    output = Path(__file__).parent.parent / "mccode_antlr" / "grammar"

    newest = grammar_file.stat().st_mtime
    for dep in deps or []:
        newest = max(newest, Path(__file__).parent.joinpath(f'{dep}.g4').stat().st_mtime)

    if language_missing_or_outdated(grammar_file, newest, features, output, **kwargs):
        rebuild_language(grammar_file, target, features, output=output, **kwargs)


def main():
    from argparse import ArgumentParser

    parser = ArgumentParser(prog="mccode-antlr-build", description='Ensure ANTLR files are up-to-date', allow_abbrev=False)
    parser.add_argument('-v', '--version', type=str, default=None, help='Version of ANTLR to build with')
    parser.add_argument('--verbose', action='store_true', help='Print out more information')
    parser.add_argument('--dryrun', action='store_true', help='Setup but do not exectute build')
    args = parser.parse_args()
    kwargs = {
        'version': args.version, 'verbose' : args.verbose, 'dryrun' : args.dryrun
    }
    mc_kwargs = {
        'target': Target.python,
        'features': [Feature.visitor],
        'deps': ('McCommon', 'c99'),
    }
    c_kwargs = {
        'target': Target.python,
        'features': [Feature.visitor, Feature.listener],
    }
    if kwargs['version'] is None:
        kwargs['version'] = antlr4_runtime_version()

    ensure_language_up_to_date('McComp', **mc_kwargs, **kwargs)
    ensure_language_up_to_date('McInstr', **mc_kwargs, **kwargs)
    ensure_language_up_to_date('C', **c_kwargs, **kwargs)

if __name__ == '__main__':
    # __name__ = 'builder.py'
    main()



from __future__ import annotations
from dataclasses import dataclass, field
from typing import TypeVar

from loguru import logger
from ..grammar import CParser, McInstrParser, CVisitor
from ..instr import InstrVisitor
from ..common import Expr

def string_between_tokens(start_token, stop_token):
    stream = start_token.getInputStream()
    return stream.getText(start_token.start, stop_token.stop)


def literal_string(ctx, until=None):
    if until is None:
        until = ctx
    return string_between_tokens(ctx.start, until.stop)


def make_error_listener(super_class, source: str, pre=5, post=2):
    class ErrorListener(super_class):
        def __init__(self):
            self.source = source
            self.pre = pre
            self.post = post

        def syntaxError(self, recognizer, offendingSymbol, *args, **kwargs):
            if len(args) == 4 and isinstance(args[3], str):
                # The 'speedy-antlr' syntax
                char_index, line, column, msg = args
            else:
                # the antlr4 (4.13.0) syntax
                line, column, msg, e = args
            logger.error(f'Syntax error in parsing {line},{column}')
            lines = self.source.split('\n')
            pre_lines = lines[line-self.pre:line]
            post_lines = lines[line:line+self.post]
            for line in pre_lines:
                logger.info(line)
            logger.error('~'*column + '^ ' + str(msg))
            for line in post_lines:
                logger.info(line)

    return ErrorListener()

TCDeclarator = TypeVar('TCDeclarator', bound='CDeclarator')
TCFuncPointer = TypeVar('TCFuncPointer', bound='CFuncPointer')

def not_both_None_or_not_equal(a, b):
    if (a is None and b is not None) or (a is not None and b is None):
        return False
    return a is not None and b is not None and a != b

def same_type(a, b):
    return type(a) == type(b)

@dataclass
class CFuncPointer:
    declare: TCDeclarator
    modifiers: str | None = None
    args: str | None = None

    @property
    def name(self) -> str:
        return self.declare.name

    def copy(self, suffix: str | None = None) -> TCFuncPointer:
        return CFuncPointer(
            declare=self.declare.copy(suffix=suffix),
            modifiers=self.modifiers,
            args=self.args,
        )

    def __eq__(self, other: TCFuncPointer) -> bool:
        if not isinstance(other, CFuncPointer):
            raise ValueError('Type mismatch')
        if not_both_None_or_not_equal(self.args, other.args):
            return False
        if not_both_None_or_not_equal(self.modifiers, other.modifiers):
            return False
        return self.declare == other.declare

    def string(self, dec_str):
        f = f'({self.modifiers} {dec_str})' if self.modifiers else f'({dec_str})'
        return f'{f}({self.args})' if self.args else f'{f}()'

    def __str__(self):
        return self.string(str(self.declare))

    def __hash__(self):
        return hash(str(self))

    def as_struct_member(self, dims: int, max_array_length):
        dec = self.declare.as_struct_member(dims=dims, max_array_length=max_array_length)
        return self.string(dec)


@dataclass
class CDeclarator:
    declare: str | CFuncPointer
    pointer: str | None = None
    extensions: list[str] = field(default_factory=list)
    elements: tuple[int,...] | tuple[str,...] | None = None
    dtype: str | None = None
    init: str | None = None

    def __post_init__(self):
        if self.elements is not None and not isinstance(self.elements, tuple):
            self.elements = tuple(self.elements)

    @property
    def is_pointer(self) -> bool:
        return self.pointer is not None and len(self.pointer.strip(' ')) > 0

    @property
    def is_array(self) -> bool:
        if self.elements is not None:
            return True
        # jump through the CFuncPointer to its CDeclarator
        return isinstance(self.declare, CFuncPointer) and self.declare.declare.is_array

    @property
    def name(self) -> str:
        return self.declare if isinstance(self.declare, str) else self.declare.name

    def copy(self, suffix: str | None = None) -> TCDeclarator:
        dec = self.declare
        if suffix is not None and isinstance(self.declare, str):
            dec = f'{self.declare}_{suffix}'
        elif suffix is not None:
            dec = self.declare.copy(suffix=suffix)
        return CDeclarator(
            pointer=self.pointer,
            declare=dec,
            extensions=[x for x in self.extensions],
            elements=self.elements,
            dtype=self.dtype,
            init=self.init,
        )

    def __eq__(self, other: TCDeclarator):
        if not isinstance(other, CDeclarator):
            raise ValueError('Type mismatch')
        if len(self.extensions) != len(other.extensions):
            return False
        for a, b in zip(self.extensions, other.extensions):
            if a != b:
                return False
        for a, b in zip((self.pointer, self.elements, self.dtype, self.init),
                        (other.pointer, other.elements, other.dtype, other.init)):
            if not_both_None_or_not_equal(a, b):
                return False
        return same_type(self.declare, other.declare) and self.declare == other.declare

    def string(self, dec_str):
        ext = " ".join(f'{x}' for x in self.extensions)
        dec = f'{dec_str} {ext}' if len(ext) else f'{dec_str}'
        if self.pointer:
            dec = f'{self.pointer} {dec}'
        if self.dtype:
            dec = f'{self.dtype} {dec}'
        return dec

    def __str__(self):
        return self.string(str(self.declare))

    def __hash__(self):
        return hash(str(self))

    def as_struct_member(self, dims: int = 0, max_array_length: int = 16384):
        if self.init:
            max_array_length = extract_initializer_size(self.init)
        dims = len(self.elements) if self.elements else 0
        mal = max_array_length if isinstance(max_array_length, tuple) else (max_array_length,)
        if len(mal) < dims:
            mal *= dims // len(mal)
        no = self.elements if self.elements else mal
        if not isinstance(no, tuple):
            raise RuntimeError(f'{no} should be a tuple but is a {type(no)}')
        if isinstance(self.declare, CFuncPointer):
            return self.string(self.declare.as_struct_member(dims=dims, max_array_length=no))
        elif self.elements is not None:
            if 0 in no:
                # Is this a bad idea?
                no = tuple(m if x == 0 else x for x, m in zip(no, mal))
            return f"{self}{''.join(f'[{n}]' for n in no)}"
        return str(self)


def extract_initializer_size(init: str) -> tuple[int,...]:
    """Assuming that the provided initializer string is for an array, possibly
    of more statically sized arrays, and that it has been written correctly,
    extract the size of each nested dimension

    >>> extract_initializer_size("{{0, 1, 2}, {3, 4, 5}}")
    (2, 3)

    >>> extract_initializer_size('{{{0}, {1}}, {{2}, {3}}, {{4}, {5}}, {{6}, {7}}}')
    (4, 2, 1)
    """
    import re
    size = []
    inner = re.compile('{([^{}]*)}')
    x = init
    while len(y:=inner.findall(x)):
        size.append(len(y[0].split(',')))
        x = inner.sub('_', x)
    return tuple(reversed(size))


class DeclaresCVisitor(CVisitor):
    def __init__(self, typedefs: list | None = None, verbose: bool = False):
        self.verbose = verbose
        self.typedefs = [x for x in typedefs] if typedefs else []
        self.declares = []

    def debug(self, message):
        if self.verbose:
            logger.debug(message)

    def info(self, message):
        if self.verbose:
            logger.info(message)

    def visitDeclaration(self, ctx:CParser.DeclarationContext):
        if ctx.staticAssertDeclaration() is not None:
            return
        #
        specs = [self.visit(x) for x in ctx.declarationSpecifiers().declarationSpecifier()]
        if ctx.initDeclaratorList() is not None:
            inits = [self.visit(x) for x in ctx.initDeclaratorList().initDeclarator()]
        else:
            inits = []

        if 'typedef' in specs:
            # this _is_ a typedef statement.
            # It (probably) is of the form
            #   typedef typeInformation aliasIdentifier;
            if specs[0] != 'typedef':
                self.debug(f"Expected typedef ... identifier; not\n{literal_string(ctx)}")
            if len(inits):
                self.debug(f'Expected no initializer with typedef, but found {inits}')
            alias = specs[-1]
            if alias in self.typedefs:
                self.info(f'Re-definition of typedef alias {alias}?')
            self.typedefs.append(alias)
        elif len(inits):
            for decl, init in inits:
                decl.dtype = ' '.join(specs)
                decl.init = init
                self.declares.append(decl)
        elif len(specs) > 1:
            decl = CDeclarator(pointer=None, declare=specs[-1], extensions=[])
            decl.dtype = ' '.join(specs[:-1])
            self.declares.append(decl)

    # Five declaration specifiers
    def visitStorageClassSpecifier(self, ctx:CParser.StorageClassSpecifierContext):
        self.debug(f'StorageClassSpecifier {literal_string(ctx)}')
        return literal_string(ctx)

    def visitTypeSpecifier(self, ctx:CParser.TypeSpecifierContext):
        if (c := ctx.atomicTypeSpecifier()) is not None:
            return self.visit(c)
        if (c := ctx.structOrUnionSpecifier()) is not None:
            return self.visit(c)
        if (c := ctx.enumSpecifier()) is not None:
            return self.visit(c)
        if (c := ctx.typedefName()) is not None:
            return self.visit(c)
        if (c := ctx.constantExpression()) is not None:
            return f'__typeof__({self.visit(c)}'
        return literal_string(ctx)

    # Keep going from here
    def visitTypeQualifier(self, ctx:CParser.TypeQualifierContext):
        self.debug(f'typeQualifier {literal_string(ctx)}')
        return literal_string(ctx)

    def visitFunctionSpecifier(self, ctx:CParser.FunctionSpecifierContext):
        self.debug(f'functionSpecifier {literal_string(ctx)}')
        return literal_string(ctx)

    def visitAlignmentSpecifier(self, ctx:CParser.AlignmentSpecifierContext):
        self.debug(f'alignmentSpecifier {literal_string(ctx)}')
        return literal_string(ctx)

    # initDeclarator
    def visitInitDeclarator(self, ctx:CParser.InitDeclaratorContext):
        # declarator ( '=' initializer )?
        decl = self.visit(ctx.declarator())
        init = self.visit(ctx.initializer()) if ctx.initializer() else None
        return decl, init

    def visitAtomicTypeSpecifier(self, ctx:CParser.AtomicTypeSpecifierContext):
        self.debug(f'AtomicTypeSpecifier {literal_string(ctx)}')
        return literal_string(ctx)

    def visitStructOrUnionSpecifier(self, ctx:CParser.StructOrUnionContext):
        self.debug(f'StructOrUnion {literal_string(ctx)}')
        return literal_string(ctx)

    def visitEnumSpecifier(self, ctx:CParser.EnumSpecifierContext):
        self.debug(f'EnumSpecifier {literal_string(ctx)}')
        return literal_string(ctx)

    def visitTypedefName(self, ctx:CParser.TypedefNameContext):
        self.debug(f'TypedefName {literal_string(ctx)}')
        return literal_string(ctx)

    def visitConstantExpression(self, ctx:CParser.ConstantExpressionContext):
        a = ctx.conditionalExpression()
        self.debug(f'ConstantExpression {literal_string(ctx)}')
        return literal_string(a)

    def visitInitializer(self, ctx:CParser.InitializerContext):
        self.debug(f'Initializer {literal_string(ctx)}')
        return literal_string(ctx)

    def visitDeclarator(self, ctx:CParser.DeclaratorContext):
        ptr = self.visit(ctx.pointer()) if ctx.pointer() else None
        dec = self.visit(ctx.directDeclarator())
        extensions = [self.visit(x) for x in ctx.gccDeclaratorExtension()]
        elements = None
        if isinstance(dec, CFuncPointer):
            # elements = dec.declare.elements
            # dec.declare.elements = None
            pass
        elif all(x in dec for x in ('[', ']')):
            dec, post = dec.split('[', 1)
            nums = tuple()
            while ']' in post:
                num, post = post.split(']', 1)
                nums += (num,)
                if '[' in post:
                    _, post = post.split('[', 1)
            try:
                elements = tuple(int(n) if len(n) else 0 for n in nums)
            except ValueError as er:
                logger.info(f"Could not convert an integer from {nums} due to {er}")
                elements = nums
        return CDeclarator(pointer=ptr, declare=dec, extensions=extensions, elements=elements)

    def visitPointer(self, ctx:CParser.PointerContext):
        self.debug(f'pointer {literal_string(ctx)}')
        return literal_string(ctx)

    def visitDirectDeclarator(self, ctx:CParser.DirectDeclaratorContext):
        if ctx.declarator():
            # this _is_ a function pointer declarator, right?
            modifiers = literal_string(ctx.vcSpecificModifer()) if ctx.vcSpecificModifer() else None
            return CFuncPointer(self.visit(ctx.declarator()), modifiers=modifiers)
        if ctx.Identifier():
            return literal_string(ctx)

        dd = ctx.directDeclarator()
        after_dd_str = literal_string(ctx).replace(literal_string(dd), '')
        dec = self.visit(dd)
        if isinstance(dec, CFuncPointer):
            dec.args = after_dd_str.strip('()')
            return dec
        return dec + after_dd_str


def extract_c_declared_variables_and_defined_types(
        block: str, user_types: list = None, verbose=False
) -> tuple[list[CDeclarator], list[str]]:
    from antlr4 import InputStream, CommonTokenStream
    from antlr4.error.ErrorListener import ErrorListener
    from ..grammar import CLexer
    stream = InputStream(block)
    lexer = CLexer(stream)
    tokens = CommonTokenStream(lexer)
    parser = CParser(tokens)
    parser.addErrorListener(make_error_listener(ErrorListener, block))
    tree = parser.compilationUnit()
    visitor = DeclaresCVisitor(user_types, verbose=verbose)
    visitor.visitCompilationUnit(tree)
    return visitor.declares, visitor.typedefs


def extract_c_declared_variables(
        block: str, user_types: list = None, verbose=False
) -> list[CDeclarator]:
    variables, _ = extract_c_declared_variables_and_defined_types(block, user_types, verbose=verbose)
    return variables

def extract_c_defined_types(
        block: str, user_types: list = None, verbose=False
) -> list[str]:
    _, types = extract_c_declared_variables_and_defined_types(block, user_types, verbose=verbose)
    return types


def extract_c_defined_then_declared_variables(
        defined_in_block: str, declared_in_block
) -> list[CDeclarator]:
    _, defined_in_types = extract_c_declared_variables_and_defined_types(defined_in_block)
    return extract_c_declared_variables(declared_in_block, user_types=defined_in_types)


class EvalCVisitor(InstrVisitor):
    def __init__(self, found: dict = None, verbose=False):
        super().__init__(None, None)
        self.verbose = verbose
        self.assigned = {} if found is None else found
        self.prog = []

    def debug(self, message):
        if self.verbose:
            logger.debug(message)

    def visitAssignment(self, ctx: McInstrParser.AssignmentContext):
        name = str(ctx.Identifier())
        value = self.visit(ctx.expr())
        self.prog.append((name, value))
        self.assigned[name] = value
        self.debug(f'{literal_string(ctx)} -> {name} = {value}')

    def visitExpressionIdentifier(self, ctx: McInstrParser.ExpressionIdentifierContext):
        from ..common import Expr
        # check if we already have a value for this identifier and insert it if we do:
        name = str(ctx.Identifier())
        if name in self.assigned:
            return self.assigned[name]
        return Expr.id(name)


def evaluate_c_defined_variables(
        variables: dict[str, str],
        initialized_in: str,
        known: dict[str, Expr] = None,
        verbose=False
):
    """Evaluate individual statements from C-like source in an attempt to find values for the provided variables"""
    from antlr4 import InputStream
    from ..grammar import McInstr_parse, McInstr_ErrorListener
    from ..common import Expr, DataType
    lines = [f'{line};' for line in initialized_in.split(';') if len(line.strip())]
    found = {} if known is None else known
    for line in lines:
        tree = McInstr_parse(InputStream(line), 'assignment', make_error_listener(McInstr_ErrorListener, line))
        visitor = EvalCVisitor(found=found, verbose=verbose)
        visitor.visitAssignment(tree)
        found = visitor.assigned

    def get_found(name, data_type):
        expr = found.get(name, Expr.id(name)).simplify()
        if expr.is_singular:
            expr.data_type = DataType.from_name(data_type)
        return expr

    return {v: get_found(v, data_type) for v, data_type in variables.items()}


def _get_expr(type_name: str, initial_value: str) -> Expr:
    from ..common import DataType, Value
    expr = Expr(Value(None)) if initial_value is None else Expr.parse(initial_value)
    expr.data_type = DataType.from_name(type_name)
    return expr


def extract_c_declared_expressions(
        block: str, user_types: list = None, verbose=False
) -> dict[CDeclarator, Expr]:
    variables = extract_c_declared_variables(block, user_types, verbose=verbose)
    return {d: _get_expr(d.dtype, d.init) for d in variables}


def evaluate_c_defined_expressions(
        variables: dict[str, Expr], initialized_in: str, verbose=False
) -> dict[str, Expr]:
    """For defined identifiers, evaluate a (simple) block of C code to determine the end values of the identifiers"""
    names_types = {name: expr.data_type.name for name, expr in variables.items()}
    return evaluate_c_defined_variables(names_types, initialized_in, known=variables, verbose=verbose)



from __future__ import annotations
from functools import cache
from unittest import TestCase
from mccode_antlr.compiler.check import simple_instr_compiles, compiled, gpu_only, mpi_only

@cache
def mcpl_config_available():
    try:
        import subprocess
        subprocess.run(['mcpl-config', '--version'], check=True)
    except FileNotFoundError:
        return False
    return True


def mcpl_compiled(method):
    def wrapper(*args, **kwargs):
        if mcpl_config_available() and simple_instr_compiles('cc'):
            method(*args, **kwargs)
        elif isinstance(args[0], TestCase):
            msg = "working compiler" if mcpl_config_available() else "mcpl interface"
            args[0].skipTest(f'Skipping due to missing {msg}')

    return wrapper


def compile_and_run(instr, parameters, run=True, dump_source=True, target: dict | None = None, config: dict | None = None):
    from pathlib import Path
    from tempfile import TemporaryDirectory
    from mccode_antlr.translators.target import MCSTAS_GENERATOR
    from mccode_antlr.run import mccode_compile, mccode_run_compiled

    kwargs = dict(generator=MCSTAS_GENERATOR, target=target, config=config, dump_source=dump_source)

    with TemporaryDirectory() as directory:
        binary, target = mccode_compile(instr, directory, **kwargs)
        # The runtime output directory used *can not* exist for McStas/McXtrace to work properly.
        # So find a name inside this directory that doesn't exist (any name should work)
        return mccode_run_compiled(binary, target, Path(directory).joinpath('t'), parameters) if run else (None, None)



from __future__ import annotations

from setuptools import Command, setup
import setuptools.command.build

from pathlib import Path
from enum import Enum

class Target(Enum):
    python = 0
    cpp = 1
    def __str__(self):
        if self == Target.python:
            return 'Python3'
        elif self == Target.cpp:
            return 'Cpp'
        raise ValueError(f'Unknown target {self}')


class Feature(Enum):
    listener = 0
    visitor = 1
    lexer = 2
    parser = 3
    def __str__(self):
        if self == Feature.listener:
            return 'Listener'
        if self == Feature.visitor:
            return 'Visitor'
        if self == Feature.lexer:
            return 'Lexer'
        if self == Feature.parser:
            return 'Parser'


def BuildANTLRCommand(source: Path, destination: str, grammars):

    def antlr4_runtime_version():
        """Retrieve the ANTLR4 version used by the available antlr4-python3-runtime"""
        from importlib import metadata
        try:
            return metadata.metadata('antlr4-python3-runtime').get('version')
        except metadata.PackageNotFoundError:
            raise RuntimeError("antlr4-python3-runtime is a build dependency!")

    def build_language(grammar_file: Path,
                       target: Target,
                       features: list[Feature],
                       output=None,
                       ):
        from subprocess import run
        from antlr4_tool_runner import initialize_paths, install_jre_and_antlr
        args = [
            f'-Dlanguage={target}',
            '-visitor' if Feature.visitor in features else '-no-visitor',
            '-listener' if Feature.listener in features else '-no-listener',
            '-o', output.resolve(),
            grammar_file.name
        ]
        print(f"Generating ANTLR {target} {' '.join(str(f) for f in features)} in {output} for {grammar_file}")
        # The following copies the implementation of antlr4_tool_runner.tool,
        # which pulls `args` from the system argv list
        initialize_paths()
        jar, java = install_jre_and_antlr(antlr4_runtime_version())
        run([java, '-cp', jar, 'org.antlr.v4.Tool'] + args, cwd=grammar_file.parent)

    class BuildANTLR(Command):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.build_lib = None
            self.editable_mode = False

        def initialize_options(self):
            """Initialize command state to defaults"""
            ...

        def finalize_options(self):
            """
            Populate the command state. This is where I traverse the directory
            tree to search for the *.ksc files to compile them later.
            The self.set_undefined_options is used to inherit the `build_lib`
            attribute from the `build_py` command.
            """
            self.set_undefined_options("build_py", ("build_lib", "build_lib"))

        def run(self):
            """
            Perform actions with side-effects, such as invoking a ksc to python compiler.
            The directory to which outputs are written depends on `editable_mode` attribute.
            When editable_mode == False, the outputs are written to directory pointed by build_lib.
            When editable_mode == True, the outputs are written in-place,
            i.e. into the directory containing the sources.
            The `run` method is not executed during sdist builds.
            """
            dest = Path(self.build_lib) / destination
            for grammar, options in grammars.items():
                build_language(source/f'{grammar}.g4',
                               target=options['target'],
                               features=options['features'],
                               output=dest)

        def get_output_mapping(self):
            """
            Return dict mapping output file paths to input file paths
            Example:
                dict = { "build/lib/output.py": "src/grammar/grammar.g4" }
            """
            files = {}
            dest = Path(self.build_lib) / destination
            for grammar, options in grammars.items():
                for feature in [Feature.lexer, Feature.parser] + options['features']:
                    files[dest / f"{grammar}{feature}.py"] = source / f"{grammar}.g4"
                    if deps := options.get("deps"):
                        for dep in deps:
                            files[dest / f"{dep}{feature}.py"] = source / f"{dep}.g4"
            return {str(k): str(v) for k, v in files.items()}

        def get_outputs(self):
            """Return list containing paths to output files (generated *.py files)"""
            files = []
            dest = Path(self.build_lib) / destination
            for grammar, options in grammars.items():
                for feature in [Feature.lexer, Feature.parser] + options['features']:
                    files.append(dest / f"{grammar}{feature}.py")
                    if deps := options.get("deps"):
                        files.extend(dest / f"{dep}{feature}.py" for dep in deps)
            return [str(file) for file in files]

        def get_source_files(self):
            """Returns list containing paths to input files (*.g4 ANTLR grammars)"""
            files = []
            for grammar, options in grammars.items():
                files.append(source / f"{grammar}.g4")
                if deps := options.get("deps"):
                    files.extend(source / f"{dep}.g4" for dep in deps)
            return [str(file) for file in files]

    return BuildANTLR


setuptools.command.build.build.sub_commands.append(("build_antlr", None))

setup(cmdclass={
    "build_antlr": BuildANTLRCommand(
        source=Path() / "src" / "grammar",  # grammar file loc relative to this file
        destination="mccode_antlr/grammar", # generated file loc in build dir
        grammars={
            'McComp': {
                'target': Target.python,
                'features': [Feature.visitor],
                'deps': ('McCommon', 'c99'),
            },
            'McInstr': {
                'target': Target.python,
                'features': [Feature.visitor],
                'deps': ('McCommon', 'c99'),
            },
            'C': {
                'target': Target.python,
                'features': [Feature.visitor, Feature.listener],
            },
        },
    )
})



from .c_listener import CDeclarator
from ..comp import Comp
from ..common import ComponentParameter


def define(a: ComponentParameter):
    return f'#define {a.name} (_comp->_parameters.{a.name})'


def undef(a: ComponentParameter):
    return f'#undef {a.name}'


def cogen_parameter_define(comp: Comp, declares: list[CDeclarator]):
    # All parameters get defines? Not just DEFINE PARAMETERS
    lines = [define(par) for pars in (comp.define, comp.setting, comp.output, declares) for par in pars]
    return '\n'.join(lines)


def cogen_parameter_undef(comp: Comp, declares: list[CDeclarator]):
    # The same parameters that were defined need to be undefined:
    lines = [undef(par) for pars in (comp.define, comp.setting, comp.output, declares) for par in pars]
    return '\n'.join(lines)



def make_assembler(name: str):
    from mccode_antlr.assembler import Assembler
    from mccode_antlr.reader import MCSTAS_REGISTRY
    return Assembler(name, registries=[MCSTAS_REGISTRY])


def parse_instr_string(instr_source: str):
    from mccode_antlr.loader import parse_mcstas_instr
    return parse_mcstas_instr(instr_source)


def cogen_finally(source, declared_parameters):
    lines = ["/* *****************************************************************************",
             f"* instrument {source.name} and components FINALLY",
             "***************************************************************************** */",
             ]

    for comp in source.component_types():
        lines.extend(cogen_comp_finally_class(comp, declared_parameters[comp.name]))

    # write the instrument main code, which calls component ones
    lines.append(f'int finally(void) {{ /* called by mccode_main for {source.name}:FINALLY */')
    lines.extend([f"#pragma acc update host(_{comp.name}_var)" for comp in source.components])
    lines.extend([
        '#pragma acc update host(_instrument_var)',
        '',
        ' siminfo_init(NULL);',
        ' save(siminfo_file); /* save data when simulation ends */'
    ])

    # insert user code from instrument definition
    if len(source.final):
        f, n = source.final[0].fn
        lines.extend([
            f'  /* Instrument {source.name} FINALLY */',
            f'  SIG_MESSAGE("[{source.name} FINALLY [{f}:{n}]");'
        ])
        for par in source.parameters:
            # ensure there's no conflict of names
            lines.append(f'  #define {par.name} (instrument->_parameters.{par.name})')
        for block in source.final:
            lines.append(block.to_c())
        for par in source.parameters:
            lines.append(f'  #undef {par.name}')

    lines.append('  /* call iteratively all components FINALLY */')
    for comp in source.components:
        if len(comp.type.final):
            lines.append(f'  class_{comp.type.name}_finally(&_{comp.name}_var);')

        for p in comp.parameters:
            if p.value.is_vector and p.value.vector_known:
                fullname = f'_{comp.name}_var._parameters.{p.name}'
                lines.append(f'  if({fullname}) free({fullname});')


    lines.extend([
        '  siminfo_close();'
        '',
        '  return(0);',
        '} /* finally */',
        ''
    ])
    return '\n'.join(lines)


def cogen_comp_finally_class(comp, declared_parameters):
    from .c_defines import cogen_parameter_define, cogen_parameter_undef
    if not len(comp.final):
        return []

    lines = [
        # f'_class_{comp.name} *class_{comp.name}_finally(_class_{comp.name} *_comp) {{',
        # Changed following https://github.com/McStasMcXtrace/McCode/commit/73a32858af1a5193bbb692da49379816cb0f9e38
        f'void class_{comp.name}_finally(_class_{comp.name} *_comp) {{',
        cogen_parameter_define(comp, declared_parameters)
    ]
    f, n = comp.final[0].fn if len(comp.final) else (comp.name, 0)
    lines.append(f'  SIG_MESSAGE("[_{comp.name}_finally] component NULL={comp.name}() [{f}:{n}]");')

    for block in comp.final:
        lines.append(block.to_c())

    lines.extend([
        cogen_parameter_undef(comp, declared_parameters),
        # '  return(_comp);',
        '  return;',
        f'}} /* class_{comp.name}_finally */',
        ''
    ])
    return lines


# def cogen_instance_finally(instance):
#     lines = [
#         f'/* component {instance.name} = {instance.type.name}() FINALLY */',
#         f'void _{instance.name}_finally(void) {{',
#     ]
#     if len(instance.type.final):
#         lines.append(f'  class_{instance.type.name}_finally(&_{instance.name}_var);')
#
#     for p in instance.parameters:
#         if p.value.is_vector and p.value.vector_known:
#             fullname = f'_{instance.name}_var._parameters.{p.name}'
#             lines.append(f'  if({fullname}) free({fullname});')
#
#     lines.extend([
#         '} /* _{instance.name}_finally */',
#         ''
#     ])
#
#     return lines



from .parameters import InstrumentParameter, ComponentParameter, parameter_name_present
from .expression import Expr, unary_expr, binary_expr, Value, UnaryOp, BinaryOp, TrinaryOp, DataType, ShapeType, ObjectType
from .metadata import MetaData
from .block import RawC, blocks_to_raw_c
from .textwrap import TextWrapper, HTMLWrapper
from .mode import Mode
from . import utilities

__all__ = [
    'parameter_name_present',
    'InstrumentParameter',
    'ComponentParameter',
    'Expr',
    'unary_expr',
    'binary_expr',
    'Value',
    'UnaryOp',
    'BinaryOp',
    'TrinaryOp',
    'DataType',
    'ShapeType',
    'ObjectType',
    'MetaData',
    'Mode',
    'RawC',
    'blocks_to_raw_c',
    'utilities',
    'TextWrapper',
    'HTMLWrapper'
]


from loguru import logger

_GETDISTANCE_FCT = """
double index_getdistance(long first_index, long second_index)
/* Calculate the distance two components from their indexes*/
{
  return coords_len(coords_sub(POS_A_COMP_INDEX(first_index), POS_A_COMP_INDEX(second_index)));
}

double getdistance(char* first_component, char* second_component)
/* Calculate the distance between two named components */
{
  int first_index = _getcomp_index(first_component);
  int second_index = _getcomp_index(second_component);
  return index_getdistance(first_index, second_index);
}

double checked_setpos_getdistance(long current_index, char* first_component, char* second_component)
/* Calculate the distance between two named components at *_setpos() time, with component index checking */
{
  int first_index = _getcomp_index(first_component);
  int second_index = _getcomp_index(second_component);
  if (first_index >= current_index || second_index >= current_index) {
    printf(\"setpos_getdistance can only be used with the names of components before the current one!\\n\");
    return 0;
  }
  return index_getdistance(first_index, second_index);
}
#define setpos_getdistance(first, second) checked_setpos_getdistance(current_setpos_index, first, second)
"""


def _split_xyz_ref(xyz_ref):
    x, y, z = [f'{v:p}' for v in xyz_ref[0]]
    return x, y, z, xyz_ref[1]


def cogen_comp_init_position(index, comp, last, instr):
    ref = None if index == 0 else instr.components[last]
    var = f'_{comp.name}_var'
    lines = [
        f'  /* component {comp.name}={comp.type.name}() AT ROTATED */',
        f'  // {comp.partial_str()}',
        '  {',
        '    Coords tc1, tc2;',
        '    tc1 = coords_set(0,0,0);',
        '    tc2 = coords_set(0,0,0);',
        '    Rotation tr1;',
        '    rot_set_rotation(tr1,0,0,0);'
    ]
    # Rotation first
    x, y, z, rel = _split_xyz_ref(comp.rotate_relative)
    if rel is None:
        # logger.debug(f'{comp.name} has absolute orientation with rotation ({x}, {y}, {z})')
        lines.append(
            f'    rot_set_rotation({var}._rotation_absolute, ({x})*DEG2RAD, ({y})*DEG2RAD, ({z})*DEG2RAD);'
        )
    else:
        # logger.debug(f'{comp.name} has relative orientation to {rel.name} with rotation ({x}, {y}, {z})')
        lines.extend([
            f'    rot_set_rotation(tr1, ({x})*DEG2RAD, ({y})*DEG2RAD, ({z})*DEG2RAD);',
            f'    rot_mul(tr1, _{rel.name}_var._rotation_absolute, {var}._rotation_absolute);'
        ])
    # Sort out relative orientation
    if index == 0:
        lines.append(f'    rot_copy({var}._rotation_relative, {var}._rotation_absolute);')
    else:
        lines.extend([
            f'    rot_transpose(_{ref.name}_var._rotation_absolute, tr1);',
            f'    rot_mul({var}._rotation_absolute, tr1, {var}._rotation_relative);'
        ])
    lines.append(f'    {var}._rotation_is_identity = rot_test_identity({var}._rotation_relative);')

    # Then translation
    x, y, z, rel = _split_xyz_ref(comp.at_relative)
    if rel is None:
        # logger.debug(f'{comp.name} has absolute positioning with rotation ({x}, {y}, {z})')
        lines.append(f'    {var}._position_absolute = coords_set({x}, {y}, {z});')
    else:
        # logger.debug(f'{comp.name} has relative positioning to {rel.name} with rotation ({x}, {y}, {z})')
        lines.extend([
            f'    tc1 = coords_set({x}, {y}, {z});',
            f'    rot_transpose(_{rel.name}_var._rotation_absolute, tr1);',
            '    tc2 = rot_apply(tr1, tc1);',
            f'    {var}._position_absolute = coords_add(_{rel.name}_var._position_absolute, tc2);'
        ])
    if index == 0:
        lines.append(f'    tc1 = coords_neg({var}._position_absolute);')
    else:
        lines.append(f'    tc1 = coords_sub(_{ref.name}_var._position_absolute, {var}._position_absolute);')
    lines.extend([
        f'    {var}._position_relative = rot_apply({var}._rotation_absolute, tc1);',
        f'  }} /* {comp.name}={comp.type.name}() AT ROTATED */',
        f'  DEBUG_COMPONENT("{comp.name}", {var}._position_absolute, {var}._rotation_absolute);',
        f'  instrument->_position_absolute[{1 + index}] = {var}._position_absolute;',
        f'  instrument->_position_relative[{1 + index}] = {var}._position_relative;',
        f'  {var}._position_relative_is_zero = coords_test_zero({var}._position_relative);'
    ])
    return '\n'.join(lines)


def cogen_comp_setpos(index, comp, last, instr, component_declared_parameters):
    from ..common import Expr

    def parameter_line(default):
        # cogen_comp_init_par skipped any parameters which had "val" which evaluates False
        # For 'out_par' (DECLARE extracted parameters) any non-'->isoptional' values were skipped.
        # For 'set_par' any `exp_tostring(entry->val)` parameters returning NULL (which was none of them?)
        p = comp.get_parameter(default.name)
        if p is None:
            return ''
        pl = []
        fullname = f'_{comp.name}_var._parameters.{p.name}'
        value = f'{p.value:p}'
        if default.value.is_str or p.value.is_str:
            # p might be an identifier, or a string literal, or an operation
            # (function call, struct member access, ...)
            # in any case we need to copy it instead of assigning
            if p.value.is_id or p.value.is_op:
                pl.extend([
                    f'  if ({value} && strlen({value})){{',
                    f'    stracpy({fullname}, {value}, 16384);',
                    '  } else {',
                    f"    {fullname}[0] = '\\0';",
                    '  }',
                ])
            elif p.value.has_value and p.value.value != '0' and p.value.value != 'NULL' and p.value.value != '""':
                # len(value)-1 to remove quotes, but copy null terminator
                pl.append(f'  stracpy({fullname}, {value}, {len(p.value.value)-1});')
            else:
                pl.append(f"  {fullname}[0] = '\\0';")
        elif default.value.is_vector or p.value.is_vector:
            if p.value.vector_known:
                # hack-in an allocator for the fixed-length allocated array
                c_type = par.value.mccode_c_type.translate(str.maketrans('', '', ' *'))  # strip the trailing ' *'
                pl.append(f'  {fullname} = calloc(sizeof({c_type}), {len(p.value.value)});')
                for i, v in enumerate(p.value.value):
                    pl.append(f'  {fullname}[{i}] = {v};')
            elif p.value.is_id:
                pl.append(f'  {fullname} = {value};')
            else:
                # it's a string (if not None), and should be an identifier that is a pointer to the data
                pl.append(f'  {fullname} = {value if p.value.has_value else "NULL"};')
        # elif p.value.is_a(Value.Type.symbol):
        #   symbol is not handled (yet) but should be initialized to zero
        elif p.value.is_id or p.value.is_op:
            pl.append(f"  {fullname} = {value if p.value is not None else 0};")
        else:
            pl.append(f'  {fullname} = {p.value if p.value.has_value else 0};')
        return '\n'.join(pl)

    f, n = comp.type.initialize[0].fn if len(comp.type.initialize) else (comp.name, 0)

    lines = [
        f'/* component {comp.name}={comp.type.name}() SETTING, POSITION/ROTATION */',
        f'int _{comp.name}_setpos(void)',
        "{ /* sets initial component parameters, position and rotation */",
        # init parameters. These can then be used in position/rotation syntax
        # all these parameters have a #define pointing to the real name space in structure
        f'  SIG_MESSAGE("[_{comp.name}_setpos] component {comp.name}={comp.type.name}() SETTING [{f}:{n}]");',
        f'  stracpy(_{comp.name}_var._name, "{comp.name}", {min(len(comp.name)+1, 16384)});',
        f'  stracpy(_{comp.name}_var._type, "{comp.type.name}", {min(len(comp.type.name)+1, 16384)});',
        f'  long current_setpos_index = _{comp.name}_var._index = {1 + index};'  # _index is a long
    ]

    # <<< This is the first call to `cogen_comp_init_par`: `cogen_comp_init_par(comp, instr, "SETTING")
    # With `section == "SETTING"` the first call picks out `comp->def->set_par`
    for par in comp.type.setting:
        # For each component definition SETTING PARAMETER the same-named instance parameter is retrieved
        lines.append(parameter_line(par))
    # >>> End of first call to `cogen_comp_init_par`
    # <<< Start of second call to `cogen_comp_init_par`: `cogen_comp_init_par(comp, instr, "PRIVATE")`
    for c_dec in component_declared_parameters[comp.type.name]:
        # c_dec is a CDeclare named tuple with (name, type, init, is_pointer, is_array, orig)
        initialized_value = 'NULL' if (c_dec.is_pointer and c_dec.init is None) else c_dec.init
        fullname = f'_{comp.name}_var._parameters.{c_dec.name}'
        if c_dec.is_array and initialized_value is not None:
            for i, val in enumerate(initialized_value.strip('{}').split(',')):
                lines.append(f'  {fullname}[{i}] = {val.strip()};')
        # elif c_dec.is_struct and initialized_value is not None:
        #     # deal with a structured initializer ...
        #     # = {.x=#, .y=#, ...}
        #     pass
        elif initialized_value is not None:
            lines.append(f'  {fullname} = {initialized_value};')
    # >>> End of second call to `cogen_comp_init_par`

    # position/rotation
    lines.append(cogen_comp_init_position(index, comp, last, instr))

    i1 = 1 + index
    lines.extend([
        f'  instrument->counter_N[{i1}] = instrument->counter_P[{i1}] = instrument->counter_P2[{i1}] = 0;',
        f'  instrument->counter_AbsorbProp[{i1}] = 0;',
        f'  return(0);',
        f'}} /* _{comp.name}_setpos */'
    ])
    return lines


def cogen_initialize(source, component_declared_parameters, ok_to_skip):
    lines = ["/* *****************************************************************************",
             f"* instrument '{source.name}' and components INITIALISE",
             "***************************************************************************** */",
             _GETDISTANCE_FCT, ""]

    last = 0
    for idx, comp in enumerate(source.components):
        lines.extend(cogen_comp_setpos(idx, comp, last, source, component_declared_parameters))
        if ok_to_skip is not None and not ok_to_skip[idx]:
            last = idx

    # generate class functions
    for comp in source.component_types():
        lines.extend(cogen_comp_initialize_class(comp, component_declared_parameters[comp.name]))

    # write the instrument main code, which calls component ones
    lines.extend([
        f'int init(void) {{ /* called by mccode_main for {source.name}:INITIALIZE */',
        '  DEBUG_INSTR();',
        '',
        '  /* code_main/parseoptions/readparams sets instrument parameters value */',
        f'  stracpy(instrument->_name, "{source.name}", {len(source.name)+1});',
    ])

    # insert user code from instrument definition
    if len(source.initialize):
        f, n = source.initialize[0].fn
        lines.extend([
            f'  /* Instrument {source.name} INITIALIZE */',
            f'  SIG_MESSAGE("[{source.name} INITIALIZE [{f}:{n}]");'
        ])
        print(f'The instrument has {len(source.parameters)} parameters')
        for par in source.parameters:
            # ensure there's no conflict of names
            lines.append(f'  #define {par.name} (instrument->_parameters.{par.name})')
        for block in source.initialize:
            lines.append(block.to_c())
        for par in source.parameters:
            lines.append(f'  #undef {par.name}')
    else:
        print("No initialization present?")

    for comp in source.components:
        lines.append(f'  _{comp.name}_setpos(); /* type {comp.type.name} */')

    lines.append('/* call iteratively all components INITIALIZE */')
    for comp in source.components:
        if len(comp.type.initialize):
            lines.append(f'  class_{comp.type.name}_initialize(&_{comp.name}_var);')

    lines.extend([
        '  if (mcdotrace) display();',
        '  DEBUG_INSTR_END();',
        '',
        # call acc_attach for OpenACC (last thing to do in init)
        '#ifdef OPENACC',
        '#include <openacc.h>',
        # update component instances on device (e.g. push managed ata structures)
        *[f'#pragma acc update device(_{comp.name}_var)' for comp in source.components],
        # update instrument
        '#pragma acc update device(_instrument_var)',
        '#endif',
        '',
        '  return(0);',
        '} /* initialize */',
        ''
    ])

    return '\n'.join(lines)


def cogen_comp_initialize_class(comp, declared_parameters):
    from .c_defines import cogen_parameter_define, cogen_parameter_undef
    if not len(comp.initialize):
        return []

    lines = [
        # f'_class_{comp.name} *class_{comp.name}_initialize(_class_{comp.name} *_comp) {{',
        # Changed following https://github.com/McStasMcXtrace/McCode/commit/73a32858af1a5193bbb692da49379816cb0f9e38
        f'void class_{comp.name}_initialize(_class_{comp.name} *_comp) {{',
        cogen_parameter_define(comp, declared_parameters)
    ]
    f, n = comp.initialize[0].fn if len(comp.initialize) else (comp.name, 0)
    lines.append(f'SIG_MESSAGE("[_{comp.name}_initialize] component NULL={comp.name}() [{f}:{n}]");')

    for block in comp.initialize:
        lines.append(block.to_c())

    lines.extend([
        cogen_parameter_undef(comp, declared_parameters),
        # '  return(_comp);',
        '  return;',
        f'}} /* class_{comp.name}_initialize */',
        ''
    ])
    return lines
